from mindustry_campaign_stats.constants import Planet, SectorNames, ItemIds
from typing import Dict, List, Union, Optional
from datetime import datetime, timezone
import dataclasses
import re


@dataclasses.dataclass
class StorageStats:
    capacity: int
    items: Dict[str, int]


@dataclasses.dataclass
class StorageAndProductionStatsMixin:
    storage: StorageStats
    rawProduction: Dict[str, float]  # Per minute
    netProduction: Dict[str, float]  # Per minute


@dataclasses.dataclass
class SectorStats(StorageAndProductionStatsMixin):
    name: str
    availability: List[str]
    imports: Dict[str, float]  # Per minute
    exports: Dict[str, float]  # Per minute


@dataclasses.dataclass
class TotalsStats(StorageAndProductionStatsMixin):
    pass


@dataclasses.dataclass
class Stats:
    date: datetime
    planet: Planet
    totals: TotalsStats
    sectors: Optional[Dict[int, SectorStats]] = None

    def to_dict(self) -> Dict:
        ret = dataclasses.asdict(self)

        ret['date'] = ret['date'].isoformat()
        ret['planet'] = ret['planet'].value

        if not ret['sectors']:
            del ret['sectors']

        return ret


class StatsBuilder:
    settings: Dict[str, Union[bool, float, int, bytes, str]]
    planet: Planet
    filter_items: Optional[List[str]] = None
    filter_sectors: Optional[List[str]] = None

    eligible_sectors: Dict[int, Dict]
    eligible_items: List[str]

    def __init__(
        self,
        settings: Dict[str, Union[bool, float, int, bytes, str]],
        planet: Planet,
        filter_items: Optional[List[str]] = None,
        filter_sectors: Optional[List[str]] = None
    ):
        self.settings = settings
        self.planet = planet
        self.filter_items = filter_items
        self.filter_sectors = filter_sectors

        self.eligible_sectors = self.get_eligible_sectors()
        self.eligible_items = self.get_eligible_items()

    def build_sectors(self) -> Dict:
        return {
            sector_id: SectorStats(
                name=SectorNames.get(self.planet).get(sector_id, str(sector_id)),
                availability=sector_info.get('resources', []),
                storage=StorageStats(
                    capacity=sector_info.get('storageCapacity', 0),
                    items=sector_info.get('items', {})
                ),
                rawProduction={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('rawProduction', {}).items() if item_id in self.eligible_items
                },
                netProduction={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('production', {}).items() if item_id in self.eligible_items
                },
                imports={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('imports', {}).items() if item_id in self.eligible_items
                },
                exports={
                    item_id: item_info.get('mean', 0) * 60 for item_id, item_info in
                    sector_info.get('export', {}).items() if item_id in self.eligible_items
                }
            ) for sector_id, sector_info in self.eligible_sectors.items()
        }

    def build_totals(self) -> TotalsStats:
        return TotalsStats(
            storage=StorageStats(
                capacity=sum([
                    sector_info.get('storageCapacity', 0) for sector_info in self.eligible_sectors.values()
                ]),
                items={
                    item_id: sum([
                        sector_info.get('items', {}).get(item_id, 0) for sector_info in self.eligible_sectors.values()
                    ]) for item_id in self.eligible_items
                }
            ),
            rawProduction={
                item_id: sum([
                    sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in
                    self.eligible_sectors.values() if sector_info.get('rawProduction', {}).get(item_id, {}).get('mean', 0)
                ]) for item_id in self.eligible_items
            },
            netProduction={
                item_id: sum([
                    sector_info.get('production', {}).get(item_id, {}).get('mean', 0) * 60 for sector_info in
                    self.eligible_sectors.values() if sector_info.get('production', {}).get(item_id, {}).get('mean', 0)
                ]) for item_id in self.eligible_items
            }
        )

    def get_eligible_sectors(self) -> Dict[int, Dict]:
        sector_name_regex = re.compile(fr'{self.planet.value}-s-(?P<number>\d+)-info')

        sectors_info = {}

        for key, value in self.settings.items():
            sector_name_match = sector_name_regex.match(key)

            if not sector_name_match:
                continue

            sector_number = int(sector_name_match.groupdict()['number'])

            if self.filter_sectors:
                sector_name = SectorNames.get(self.planet).get(sector_number, str(sector_number)).lower()

                if not any([name for name in self.filter_sectors if name.lower() in sector_name]):
                    continue

            sectors_info[sector_number] = value

        return sectors_info

    def get_eligible_items(self) -> List[str]:
        if not self.filter_items:
            return ItemIds.get(self.planet)

        return [
            item_id for item_id in ItemIds.get(self.planet) if any([name for name in self.filter_items if name.lower() in item_id.lower()])
        ]


def compute(
        settings: Dict[str, Union[bool, float, int, bytes, str]],
        planet: Planet,
        totals_only: bool = False,
        filter_items: Optional[List[str]] = None,
        filter_sectors: Optional[List[str]] = None
) -> Stats:
    builder = StatsBuilder(settings, planet, filter_items, filter_sectors)

    return Stats(
        date=datetime.now(timezone.utc),
        planet=planet,
        totals=builder.build_totals(),
        sectors=builder.build_sectors() if not totals_only else None
    )



from mindustry_campaign_stats.constants import ItemIds, ItemColors
from mindustry_campaign_stats.stats import Stats
from math import log10, floor
from rich.table import Table
import json


def humanize_number(value, significant_digits=3, strip_trailing_zeros=True):
    powers = [10 ** x for x in (12, 9, 6, 3, 0, -3, -6, -9)]
    human_powers = ['T', 'B', 'M', 'k', '', 'm', 'µ', 'n']
    is_negative = False
    suffix = ''

    if not isinstance(value, float):
        value = float(value)
    if value < 0:
        is_negative = True
        value = abs(value)
    if value == 0:
        decimal_places = max(0, significant_digits - 1)
    elif .001 <= value < 1:
        decimal_places = max(0, significant_digits - int(floor(log10(value))) - 1)
    else:
        p = next((x for x in powers if value >= x), 10 ** -9)
        i = powers.index(p)
        value = value / p
        before = int(log10(value)) + 1
        decimal_places = max(0, significant_digits - before)
        suffix = ' ' + human_powers[i]

    return_value = ("%." + str(decimal_places) + "f") % value

    if is_negative:
        return_value = "-" + return_value
    if strip_trailing_zeros and '.' in return_value:
        return_value = return_value.rstrip('0').rstrip('.')

    return f'{return_value}{suffix}'


def to_table(computed_stats: Stats, compact: bool = False) -> Table:
    date = computed_stats.date.astimezone().strftime('%c')

    ret = Table(
        title=f'{date} - {computed_stats.planet.name}'
    )

    item_ids = computed_stats.totals.storage.items.keys()

    if not computed_stats.sectors: # Totals only mode
        # Header
        ret.add_column('Item')
        ret.add_column('Storage')
        ret.add_column('Raw production (/m)')
        ret.add_column('Net production (/m)')

        for item_id in item_ids:
            if compact and computed_stats.totals.storage.items.get(item_id, 0) == 0 and computed_stats.totals.rawProduction.get(item_id, 0) == 0:
                continue

            row = [
                item_id.replace('-', ' ').title(),
                humanize_number(computed_stats.totals.storage.items.get(item_id, 0)),
                humanize_number(computed_stats.totals.rawProduction.get(item_id, 0)),
                humanize_number(computed_stats.totals.netProduction.get(item_id, 0)),
            ]

            ret.add_row(
                *row,
                style=ItemColors.get(item_id),
                end_section=True
            )
    else: # Normal mode
        ret.show_footer = True

        # Header
        ret.add_column(
            'Sector',
            footer='Totals'
        )

        ret.add_column(
            'Stat',
            footer=f'Storage ({humanize_number(computed_stats.totals.storage.capacity)})\nRaw prod. (/m)\nNet prod. (/m)',
            no_wrap=True
        )

        sector_item_ids = []

        for item_id in item_ids:
            if compact and computed_stats.totals.storage.items.get(item_id, 0) == 0 and computed_stats.totals.rawProduction.get(item_id, 0) == 0:
                continue

            ret.add_column(
                item_id.replace('-', '\n').title(),
                footer='\n'.join([
                    humanize_number(computed_stats.totals.storage.items.get(item_id, 0)),
                    humanize_number(computed_stats.totals.rawProduction.get(item_id, 0)),
                    humanize_number(computed_stats.totals.netProduction.get(item_id, 0)),
                ]),
                style=ItemColors.get(item_id),
                header_style=ItemColors.get(item_id),
                footer_style=ItemColors.get(item_id),
                no_wrap=True
            )

            sector_item_ids.append(item_id)

        # Body
        for sector in sorted(computed_stats.sectors.values(), key=lambda sector: sector.name):
            if compact:
                items_unavailable_anywhere = len([
                    item_id for item_id in sector_item_ids if item_id not in sector.availability
                ]) == len(sector_item_ids)

                items_storage_empty = sum([
                    sector.storage.items.get(item_id, 0) for item_id in sector_item_ids
                ]) == 0

                no_raw_production = sum([
                    sector.rawProduction.get(item_id, 0) for item_id in sector_item_ids
                ]) == 0

                no_net_production = sum([
                    sector.netProduction.get(item_id, 0) for item_id in sector_item_ids
                ]) == 0

                if items_unavailable_anywhere and items_storage_empty and no_raw_production and no_net_production:
                    continue

            stat_labels_cell = [
                'Available',
                f'Storage ({humanize_number(sector.storage.capacity)})',
                'Raw prod. (/m)',
                'Net prod. (/m)',
            ]

            if sector.imports:
                stat_labels_cell.append('Imports (/m)')

            if sector.exports:
                stat_labels_cell.append('Exports (/m)')

            row = [
                str(sector.name).replace(' ', '\n'),
                '\n'.join(stat_labels_cell)
            ]

            for item_id in sector_item_ids:
                stat_values_cell = [
                    '[green]✓[/green]' if item_id in sector.availability else '[red]✕[/red]',
                    humanize_number(sector.storage.items.get(item_id, 0)),
                    humanize_number(sector.rawProduction.get(item_id, 0)),
                    humanize_number(sector.netProduction.get(item_id, 0)),
                ]

                if item_id in sector.imports:
                    stat_values_cell.append(humanize_number(sector.imports.get(item_id, 0)))

                if item_id in sector.exports:
                    stat_values_cell.append(humanize_number(sector.exports.get(item_id, 0)))

                row.append(
                    '\n'.join(stat_values_cell)
                )

            ret.add_row(
                *row,
                end_section=True
            )

    return ret


def to_json(computed_stats: Stats, pretty: bool) -> str:
    return json.dumps(
        computed_stats.to_dict(),
        indent=2 if pretty else None,
        separators=None if pretty else (',', ':')
    )



from watchdog.events import PatternMatchingEventHandler, FileModifiedEvent
from mindustry_campaign_stats.presenters import to_json, to_table
from mindustry_campaign_stats.__version__ import __version__
from mindustry_campaign_stats.constants import Planet
from mindustry_campaign_stats.settings import load
from mindustry_campaign_stats.stats import compute
from argparse import ArgumentParser, Namespace
from watchdog.observers import Observer
from rich.console import Console
from threading import Timer
from time import sleep
import os

console = Console()


def show(args: Namespace) -> None:
    if not args.json and args.refresh:
        console.clear()

    with open(args.filename, 'rb') as fp:
        settings_parsed = load(fp)

    computed_stats = compute(
        settings_parsed,
        args.planet,
        args.totals,
        args.items.split(',') if args.items else None,
        args.sectors.split(',') if args.sectors else None
    )

    if args.json:
        console.out(
            to_json(computed_stats, args.pretty if not args.refresh else False),
            end='\n'
        )
    else:
        console.print(
            to_table(computed_stats, args.compact)
        )


def cli() -> None:
    arg_parser = ArgumentParser(
        description='CLI tool to read Mindustry\'s campaign global stats.'
    )

    arg_parser.add_argument(
        'filename',
        help='The settings.bin file to load'
    )

    arg_parser.add_argument(
        'planet',
        help='Which campaign to retrieve stats for',
        type=Planet,
        choices=list(Planet)
    )

    arg_parser.add_argument(
        '-v', '--version',
        action='version',
        version=f'mindustry-campaign-stats {__version__}'
    )

    arg_parser.add_argument(
        '-j', '--json',
        help='Output JSON instead of a table',
        action='store_true'
    )

    arg_parser.add_argument(
        '-p', '--pretty',
        help='Pretty-print JSON output',
        action='store_true'
    )

    arg_parser.add_argument(
        '-r', '--refresh',
        help='Listen for file changes',
        action='store_true'
    )

    arg_parser.add_argument(
        '-t', '--totals',
        help='Compute totals only',
        action='store_true'
    )

    arg_parser.add_argument(
        '-i', '--items',
        help='Get stats for these items only (comma-separated list of item names)'
    )

    arg_parser.add_argument(
        '-s', '--sectors',
        help='Get stats for these sectors only (comma-separated list of sector names)'
    )

    arg_parser.add_argument(
        '-c', '--compact',
        help='Hide empty and useless stats',
        action='store_true'
    )

    args = arg_parser.parse_args()

    show(args)

    if args.refresh:
        class SettingsModifiedHandler(PatternMatchingEventHandler):
            def __init__(self, *args, **kvargs):
                super().__init__(*args, **kvargs)

                self.timer = None

            def on_modified(self, event: FileModifiedEvent):
                if self.timer:
                    self.timer.cancel()

                self.timer = Timer(2.0, lambda args: show(args), args=[args])
                self.timer.start()

        observer = Observer()

        observer.schedule(
            SettingsModifiedHandler(patterns=[os.path.basename(args.filename)], ignore_directories=True),
            os.path.dirname(args.filename) or './',
            recursive=False
        )

        observer.start()

        try:
            while True:
                sleep(1)
        except KeyboardInterrupt:
            observer.stop()

        observer.join()
